import base64
import binascii
import json
from typing import Optional

from django.conf import settings
from django.http import HttpRequest

from nkunyim_util.models.session_model import XanModel
from nkunyim_util.encryption.rsa_encryption import RSAEncryption



class XanAuthenticationMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request: HttpRequest):
        self.process_request(request)
        return self.get_response(request)

    def _get_xan(self, request: HttpRequest) -> Optional[XanModel]:
        header = request.META.get('HTTP_XAN_AUTHORIZATION')
        if not header:
            return None

        try:
            # Strip prefix/suffix or encoding markers
            cipher_token = header[2:-1]
            cipher_text = base64.b64decode(cipher_token)

            # Decrypt token
            plain_text = RSAEncryption().decrypt(cipher_text)
            xan_data = json.loads(plain_text)
            xan_model = XanModel(**xan_data)

            if not bool(xan_model.service_name and xan_model.service_name == settings.NKUNYIM_SERVICE):
                return None

            return xan_model
        except (ValueError, KeyError, json.JSONDecodeError, binascii.Error):
            return None


    def process_request(self, request: HttpRequest):
        xan_model = self._get_xan(request)
        if xan_model:
            xan_model.is_authenticated = True
            request.xan = xan_model # type: ignore
        else:
            request.xan = None # type: ignore



class MultipleProxyMiddleware:
    FORWARDED_FOR_FIELDS = [
        "HTTP_X_FORWARDED_FOR",
        "HTTP_X_FORWARDED_HOST",
        "HTTP_X_FORWARDED_SERVER",
    ]

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        """
        Rewrites the proxy headers so that only the most
        recent proxy is used.
        """
        for field in self.FORWARDED_FOR_FIELDS:
            if field in request.META:
                if "," in request.META[field]:
                    parts = request.META[field].split(",")
                    request.META[field] = parts[-1].strip()
        return self.get_response(request)
    
    


from rest_framework.permissions import BasePermission



class IsXanAuthenticated(BasePermission):
    """
    Allows access only to xan_authenticated users.
    """

    def has_permission(self, request, view): # type: ignore
        return bool(request.xan and request.xan.is_authenticated)






from abc import ABC, abstractmethod
from datetime import datetime
import importlib.util
import os
from pathlib import Path
import re
import shutil
from typing import Union


from django.conf import settings


PYTHON_FILE_NAME_EXT = ".py"

DATAX_NAME_PREFIX = "Datax_"

DATAX_INPUT_ERROR = "DATAX_INPUT_ERROR"
DATAX_FILE_ERROR = "DATAX_FILE_ERROR"

DATAX_MIGRATE_EXCEPTION = "DATAX_MIGRATE_EXCEPTION"
DATAX_REVERSE_EXCEPTION = "DATAX_REVERSE_EXCEPTION"

DATAX_CLASS_NAME = "DataxMigration"
DATAX_METHOD_MIGRATE = "migrate"
DATAX_METHOD_REVERSE = "reverse"


class DataxBaseCommand(ABC):
    
    def __init__(self, file_name: str) -> None:

        self.base_path = os.path.join(settings.BASE_DIR, 'nkunyim_data/migration')
        
        self.python_file_name = f"{file_name}{PYTHON_FILE_NAME_EXT}"
        
        self.pyhton_file_path = os.path.join(self.base_path, self.python_file_name)
        
        # Create Datax root directory if not exist.
        os.makedirs(self.base_path, exist_ok=True)

        template_dir = Path(__file__).parent.resolve()
        self.template_path =  os.path.join(template_dir, "template.py")


    @abstractmethod
    def go(self) -> Union[str, None]:
        pass

        
    def run(self, method_name: str) -> Union[str, None]:
        try:

            if not os.path.isfile(self.pyhton_file_path):
                return f"{self.python_file_name} is not a valid Datax Migration file."
            
            # Load module spec from file path
            module_name = os.path.splitext(os.path.basename(self.pyhton_file_path))[0]
            spec = importlib.util.spec_from_file_location(module_name, self.pyhton_file_path)

            if spec is None or spec.loader is None:
                return f"Could not load module from {self.python_file_name}"

            # Create a module and execute it
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            # Get the class from the module
            klass = getattr(module, DATAX_CLASS_NAME)
            instance = klass()

            # Get the method and call it
            method = getattr(instance, method_name)
            result = method()

            return result
            
        except Exception as e:
            return f"{DATAX_FILE_ERROR}: {e}."
        


class DataxNewCommand(DataxBaseCommand):
    
    def __init__(self, input_name: str) -> None:
        name = re.sub(r'[^a-zA-Z]', ' ', input_name)
        name = re.sub(r'([a-z])([A-Z])', r'\1 \2', name)
        pins = name.split()
        cmd_name = ''.join(pin.capitalize() for pin in pins)
        date_str = datetime.now().strftime("%Y%m%d%H%M%S")
        file_name = f"{DATAX_NAME_PREFIX}{date_str}_{cmd_name}"
        
        super().__init__(file_name=file_name)

    
    def go(self) -> Union[str, None]:
        try:
            # Copy template to implementing project and rename to user input_text 
            shutil.copy2(self.template_path, self.pyhton_file_path)
            
            return None
        except FileNotFoundError:
            return f"{DATAX_FILE_ERROR}: Source file '{self.template_path}' not found."
        except Exception as e:
            return f"{DATAX_FILE_ERROR}: Exception {e}."
        
        
        
class DataxRunCommand(DataxBaseCommand):
    
    def __init__(self, input_name: str) -> None:
        
        if not input_name.startswith(DATAX_NAME_PREFIX):
            input_name = f"{DATAX_NAME_PREFIX}{input_name}"
    
        if input_name.endswith(PYTHON_FILE_NAME_EXT):
            input_name = input_name.split(".")[0]

        super().__init__(file_name=input_name)

            
  
    def go(self) -> Union[str, None]:
        return self.run(method_name=DATAX_METHOD_MIGRATE)

        
        
class DataxRevCommand(DataxBaseCommand):
    
    def __init__(self, input_name: str) -> None:
        
        if not input_name.startswith(DATAX_NAME_PREFIX):
            input_name = f"{DATAX_NAME_PREFIX}{input_name}"
    
        if not input_name.endswith(PYTHON_FILE_NAME_EXT):
            input_name = f"{input_name}{PYTHON_FILE_NAME_EXT}"

        super().__init__(file_name=input_name)

            
  
    def go(self) -> Union[str, None]:
        return self.run(method_name=DATAX_METHOD_REVERSE)




from decimal import Decimal
from typing import Optional
from pydantic import BaseModel


class LocationModel(BaseModel):
    user_ip: str
    accuracy_radius: int
    city: str
    continent_code: str
    continent_name: str 
    country_code: str 
    country_name: str 
    is_in_european_union: bool
    latitude: Decimal
    longitude: Decimal
    metro_code: Optional[str]
    postal_code: Optional[str] 
    region_code: str
    region_name: str
    time_zone: str
    dma_code: Optional[str] 
    region: Optional[str]



from typing import Union
from django.conf import settings
from django.core.cache import DEFAULT_CACHE_ALIAS

from nkunyim_util.models.app_model import AppModel

from .cache_manager import CacheManager, DEFAULT_CACHE_TIMEOUT


    
class AppCache:

    def __init__(self, key: str) -> None:
        self.key = key
        self.cache = CacheManager(settings.APPLICATION_CACHE if hasattr(settings, 'APPLICATION_CACHE') else DEFAULT_CACHE_ALIAS)
        
    def set(self, model: AppModel, timeout: int = DEFAULT_CACHE_TIMEOUT):
        self.cache.set(key=self.key, value=model, timeout=timeout)
        
    def get(self) -> Union[AppModel, None]:
        return self.cache.get(key=self.key)
    
    def delete(self) -> None:
        self.cache.delete(key=self.key)
        
    def clear(self) -> None:
        self.cache.clear()

   


from typing import Optional, Union
from uuid import UUID
from pydantic import BaseModel

from .app_model import AccountModel
from .location_model import LocationModel
from .nation_model import NationModel
from .access_model import NavModel
from .page_model import PageModel
from .user_model import UserModel


class ContextModel(BaseModel):
    id: UUID
    account: AccountModel
    mode: str
    name: str
    title: str
    caption: str
    description: str
    keywords: str
    image_url: str
    logo_url: str
    logo_light_url: str
    logo_dark_url: str
    icon_url: str
    icon_light_url: str
    icon_dark_url: str
    tags: str
    colour: str
    is_active: bool
    location: LocationModel
    nation: NationModel
    user: Optional[UserModel]
    page: PageModel
    uix: dict
    navs: Union[list[NavModel], None]
    root: str
    
    


from datetime import datetime
from pydantic import BaseModel
from uuid import UUID


class AccountModel(BaseModel):
    id: UUID
    name: str
    title: str
    caption: str
    logo_url: str
    is_active: bool
    created_at: datetime
    updated_at: datetime
    
    
class AppModel(BaseModel):
    id: UUID
    account: AccountModel
    mode: str
    name: str
    title: str
    caption: str
    description: str
    keywords: str
    image_url: str
    logo_url: str
    logo_light_url: str
    logo_dark_url: str
    icon_url: str
    icon_light_url: str
    icon_dark_url: str
    client_id: str
    client_secret: str
    redirect_url: str
    response_type: str
    grant_type: str
    domain: str
    scope: str
    aes_key: str
    rsa_public_pem: str
    rsa_private_pem: str
    rsa_passphrase: str
    algorithm: str
    claims: dict
    tags: str
    colour: str
    is_active: bool
    created_at: datetime
    updated_at: datetime




import base64
import json
from typing import Optional
from django.conf import settings
from django.http import HttpRequest

from nkunyim_util.models.user_model import UserModel
from nkunyim_util.encryption.rsa_encryption import RSAEncryption
from nkunyim_util.http.http_client import HttpClient


class UserService:
    
    def get(self, req: HttpRequest) -> Optional[UserModel]:      
        try:
            user = req.user
            if not user.is_authenticated:
                return None
            
            client = HttpClient(req=req, name=settings.IDENTITY_SERVICE)
            response = client.get(path=f"/api/users/me")
            if not response.ok:
                # log error
                return None
            
            json_data = response.json()
            result_data = dict(json_data)
            
            return UserModel(**result_data)
        except:
            return None 
        
        
    def set(self, token: str) -> Optional[UserModel]:
        try:
            cipher_token = token[2:-1] # Cater for bytes str concatenation issue
            encryption = RSAEncryption()
            cipher_text = base64.b64decode(cipher_token)
            plain_text = encryption.decrypt(cipher_text=cipher_text)
            userinfo = json.loads(plain_text)
            user_data = dict(userinfo)
            return UserModel(**user_data)
        except Exception as ex:
            return None



from typing import Optional
from django.conf import settings
from django.http import HttpRequest

from nkunyim_util.models.app_model import AppModel
from nkunyim_util.caches.app_cache import AppCache
from nkunyim_util.http.http_client import HttpClient


from .signals import app_data_updated


class AppService:
    
    def __init__(self, req: HttpRequest, session_key: str) -> None:
        self.cache = AppCache(key=f"app.{session_key}")
        self.req = req


    def _get_from_api(self) -> Optional[AppModel]:
        try:
            domain = '.'.join(self.req.get_host().rsplit('.', 2)[-2:]).lower()
            client = HttpClient(req=self.req, name=settings.MARKET_SERVICE)
            response = client.get(path=f"/api/applications/?domain={domain}")
            if not response.ok:
                # log error
                return None
            
            json_data = response.json()
            result_data = dict(json_data).get("data", [None])
            model_data = result_data[0]

            return model_data
        except:
            return None


    def _make(self) -> AppModel:
        model_data = self._get_from_api()
        if not model_data:
            model_data = dict(settings.NKUNYIM_DEFAULT_APP)
    
        app_model = AppModel(**model_data) # type: ignore
        self.cache.set(model=app_model, timeout=60 * 60 * 24)
        
        # Inform interested parties
        app_data_updated.send(sender=AppModel, instance=app_model)
        
        return app_model
    
    
    def get(self) -> AppModel:
        return self.cache.get() or self._make()
    


from collections import defaultdict
from typing import List, Optional
from uuid import UUID

from django.conf import settings
from django.http import HttpRequest

from nkunyim_util.caches.access_cache import AccessCache
from nkunyim_util.models.access_model import AccessModel, MenuModel, NavModel
from nkunyim_util.http.http_client import HttpClient


from .signals import access_data_updated


class AccessService:
    
    def __init__(self, req: HttpRequest, session_key: str, application_id: UUID) -> None:
        self.cache = AccessCache(key=f"acc.{session_key}")
        self.application_id = str(application_id)
        self.req = req


    def _build_from_menus(self, menus: List[dict]) -> tuple[dict[str, str], List[NavModel]] :
        env = "dev"
        uix = {}
        menus_dict = defaultdict(List[MenuModel])

        for menu in menus:
            node = menu['node']
            module = menu['module']
            menu_data = {
                '_id': menu['id'],
                'node': node,
                'sequence': menu['sequence'],
                **{key: module[key] for key in (
                    'id', 'name', 'title', 'caption', 'icon', 'path', 'route', 'colour', 'tags'
                )}
            }
            items = menu['items']

            module_name = str(module['name']).title()
            module_path = str(module['path']).lower()
            
            uix[f"{module_name}Page"] = f"./{module_path}/home.{env}"

            menu_data['items'] = []
            for item in items:
                page = item['page']
                uix[f"{module_name}{str(page['name']).title()}Page"] = f"./{module_path}/{str(page['path']).lower()}.{env}"
                item_data = {
                    '_id': menu['id'],
                    'sequence': menu['sequence'],
                    **{key: page[key] for key in (
                        'id', 'name', 'title', 'caption', 'icon', 'path', 'route', 'tags'
                    )}
                }
                menu_data['items'].append(item_data)
                
            if node in {"toolbox", "manage", "system"}:
                menus_dict[node].append(MenuModel(**menu_data))

        navs = [
            NavModel(node=str(nodex).title(), menus=menux)
            for nodex, menux in menus_dict.items()
        ]

        return uix, navs
        
        
    def _get_from_api(self) -> Optional[tuple[Optional[dict], Optional[List[dict]]]]:
        try:
            client = HttpClient(req=self.req, name=settings.MARKET_SERVICE)
            response = client.get(path=f"/api/accesses/?application_id={self.application_id}")
            access = dict(response.json()) if response.ok else {}
            
            role = access.get("role", None)
            menus = access.get("menus", None) 
            return role, menus
        except:
            return None
        
        
    def _make(self) -> AccessModel:
        navs = None
        role = None
        uix = dict(settings.NKUNYIM_UIX)
        api = self._get_from_api()
        if api:
            role, menus = api
            if menus:
                navs_uix = self._build_from_menus(menus=menus)
                nui, navs = navs_uix
                uix.update(nui)
                
        model_data = {
            'role': role,
            'navs': navs,
            'uix': uix
        }
        access_model = AccessModel(**model_data)
        self.cache.set(model=access_model, timeout=60 * 60 * 24)
        
        # Inform interested parties
        access_data_updated.send(sender=AccessModel, instance=access_model)
        
        return access_model
            
    
    def get(self, refresh: bool = False) -> AccessModel:
        return self._make() if refresh else self.cache.get() or self._make()




from cryptography.fernet import Fernet


class AESEncryption:
    
    def get_key(self) -> str:
        key = Fernet.generate_key()
        return key.decode()

    
    def encrypt(self, key: str, plain_text: str) -> str:
        cipher_suite = Fernet(key.encode())
        cipher_text = cipher_suite.encrypt(plain_text.encode())
        return cipher_text.decode()

    
    def decrypt(self, key: str, cipher_text: str) -> str:
        cipher_suite = Fernet(key.encode())
        plain_text = cipher_suite.decrypt(cipher_text.encode())
        return plain_text.decode()



from typing import Optional
from django.http import HttpRequest

from nkunyim_util.models.session_model import SessionModel



class HttpSession:
    def __init__(self, req: HttpRequest) -> None:
        self.req = req
    
    def _key(self) -> str:
        parts = self.req.get_host().lower().split('.')
        if len(parts) >= 2:
            domain = '.'.join(parts[-2:])
            subdomain = parts[-3] if len(parts) > 2 else "www"
            return f"{subdomain}.{domain}"
        return "www.localhost"
    
    def set(self, model: SessionModel) -> None:
        self.req.session[self._key()] = model.model_dump()
        self.req.session.modified = True
        
    def get(self) -> Optional[SessionModel]:
        key = self._key()
        if not bool(key in self.req.session):
            return None

        session = self.req.session[key]
        return SessionModel(**session)
    
    def kill(self) -> None:
        self.req.session.clear()
        self.req.session.flush()
        




from datetime import datetime
from pydantic import BaseModel
from typing import Optional


class XanModel(BaseModel):
    client_id: str
    nation_code: str
    service_name: Optional[str]
    is_authenticated: bool
    
    
class TokenModel(BaseModel):
    access_token: str
    token_type: str
    expires_at: datetime


class SessionModel(BaseModel):
    token: Optional[TokenModel]
    xan: Optional[XanModel]